
"""Module computing the best coefficient of a QC automatically"""

import importlib.resources
import json

from pybsrnqc import coef_study as cs
from pybsrnqc import open_data as od
from pybsrnqc import plot_limits as pl
from pybsrnqc import qc_functions as qcf
from pybsrnqc.config import Coef

# Coefficients initialisation
# Get data conf from JSON file
coef = Coef()

def compute(path: str, bw_sel: str = None):

    qc = None
    while qc not in {"QC1", "QC2", "QC3", "QC5", "QC10"}:
        qc = input("What QC do you want to study ? (QC1/QC2/QC3/QC5/QC10) :")

    level = None
    while level not in {"level_1", "level_2"}:
        level = input("What level do you want to study ? (level_1/level_2) :")

    if qc == 'QC1':
        qc = qcf.QC1()
    if qc == 'QC2':
        qc = qcf.QC2()
    if qc == 'QC3':
        qc = qcf.QC3()
    if qc == 'QC5':
        qc = qcf.QC5()
    if qc == 'QC10':
        qc = qcf.QC10()

    # Get datas
    df = od.open_all(path=path + '/')

    # Calculating kernel density for the dataset
    if qc.name in {'QC3', 'QC10'}:
        log_kernel, selected = pl.kde_computing(df, qc, select=True, bw_sel=bw_sel)
    else:
        log_kernel = pl.kde_computing(df, qc, bw_sel=bw_sel)

    # Choosing the threshold
    threshold = float(input('Threshold for outliers : '))

    # Nb of coefficients tried
    nb_try = float(input("Number of coefficients tried (200 is ok): "))

    # Field of research
    coef_range = qc.coef_range
    step = (coef_range[1] - coef_range[0]) / nb_try

    if qc.vary == 'downward_avg':
        coef_range_min = qc.coef_range_min
        step_min = (coef_range_min[1] - coef_range_min[0]) / nb_try

    # Finding the best coefficient for a density threshold given
    if qc.vary == 'downward_avg':
        if qc.name == 'QC10':
            df_score, score, df_score_min, score_min = cs.calc_coef(
                df, log_kernel, qc, threshold, level=level,
                coef_range=coef_range, coef_range_min=coef_range_min,
                step=step, step_min=step_min, selected=selected)

        else:
            df_score, score, df_score_min, score_min = cs.calc_coef(
                df, log_kernel, qc, threshold, level=level,
                coef_range=coef_range, coef_range_min=coef_range_min,
                step=step, step_min=step_min)

    elif qc.vary == 'direct_avg':
        df_score, score = cs.calc_coef(df, log_kernel, qc, threshold, level=level, coef_range=coef_range, step=step,
                                       selected=selected)
        print('Best coefficient:')
        print(score)
    else:
        df_score, score = cs.calc_coef(df, log_kernel, qc, threshold, level=level, coef_range=coef_range, step=step)
        print('Best coefficient:')
        print(score)

    # In the case there are several best scores
    if score.shape != (1, 5):
        score = score.iloc[0]
    if qc.vary == 'downward_avg':
        score_min = score_min.iloc[0]

    # Charging the new coeff
    coef.__setattr__(qc.coefficients[level], float(score[qc.coefficients[level]]))
    if qc.vary == 'downward_avg':
        coef.__setattr__(qc.coefficients[level + '_min'], float(score_min[qc.coefficients[level + '_min']]))

    # Plotting the result
    pl.plot_kde(df, log_kernel, qc, coef, level=level)

    if qc.vary == 'downward_avg':
        return qc.coefficients[level], float(score[qc.coefficients[level]]), qc.coefficients[level + '_min'], float(score_min[qc.coefficients[level + '_min']])
    else:
        return qc.coefficients[level], float(score[qc.coefficients[level]])



"""
Module that provides the QC class with their assiocated functions and labeling
"""

import math

import numpy as np

import pybsrnqc.qcrad as qcr
from pybsrnqc.config import Coef


def gen_calc_lim(self, df, coef: Coef):
    """ Computation of the limits and the labels of a dataframe"""

    df = df[['timestamp', self.varx, self.vary]]

    X_val1 = np.array(df[self.varx])
    X_val2 = np.array(df[self.vary])

    lim_l1 = []
    lim_l2 = []
    lim_bsrn = []

    lim_l1_min = []
    lim_l2_min = []
    lim_bsrn_min = []

    labels = []

    for VAL1, VAL2 in zip(X_val1, X_val2):

        if self.vary == 'downward_avg':

            l1, l2, l_bsrn, l1_min, l2_min, l_bsrn_min = self.f(VAL2, VAL1, coef)

            lim_l1_min.append(l1_min)
            lim_l2_min.append(l2_min)
            lim_bsrn_min.append(l_bsrn_min)

        else:

            l1, l2, l_bsrn = self.f(VAL2, VAL1, coef)

        label = self.lab(VAL2, VAL1, coef)

        lim_l1.append(l1)
        lim_l2.append(l2)
        lim_bsrn.append(l_bsrn)
        labels.append(label)

    if self.vary == 'downward_avg':
        return lim_l1, lim_l2, lim_bsrn, labels, lim_l1_min, lim_l2_min, lim_bsrn_min
    else:
        return lim_l1, lim_l2, lim_bsrn, labels

# Definition of the QC class


class QC1:

    def __init__(self):
        self.name = 'QC1'
        self.varx = 'SZA'
        self.vary = 'global2_avg'
        self.unitx = 'SZA'
        self.unity = 'W.m-2'
        self.coefficients = {'level_1': 'C1', 'level_2': 'D1'}
        self.coef_range = [0.0, 1.2]

    @staticmethod
    def f(GSW, SZA, coef: Coef):
        ''' Return the 2 main variables, the 1rst level limit, the 2nd level limit and the physical limit for a sample'''

        l1 = qcr.REF.SOLAR_CONSTANT * coef.C1 * math.pow(math.cos(math.radians(SZA)), 1.2) + 50
        l2 = qcr.REF.SOLAR_CONSTANT * coef.D1 * math.pow(math.cos(math.radians(SZA)), 1.2) + 55
        l_bsrn = qcr.REF.SOLAR_CONSTANT * 1.5 * math.pow(math.cos(math.radians(SZA)), 1.2) + 100

        return l1, l2, l_bsrn

    @staticmethod
    def lab(GSW, SZA, coef: Coef):
        return qcr.QC1(GSW, SZA, coef)

    def calc_lim(self, df, coef: Coef):
        return gen_calc_lim(self, df, coef)


class QC2:

    def __init__(self):
        self.name = 'QC2'
        self.varx = 'SZA'
        self.vary = 'diffuse_avg'
        self.unitx = 'SZA'
        self.unity = 'W.m-2'
        self.coefficients = {'level_1': 'C2', 'level_2': 'D2'}
        self.coef_range = [0.0, 1.0]

    @staticmethod
    def f(Dif, SZA, coef: Coef):
        ''' Return the 2 main variables, the 1rst level limit, the 2nd level limit and the physical limit for a sample'''

        l1 = qcr.REF.SOLAR_CONSTANT * coef.C2 * math.pow(math.cos(math.radians(SZA)), 1.2) + 30
        l2 = qcr.REF.SOLAR_CONSTANT * coef.D2 * math.pow(math.cos(math.radians(SZA)), 1.2) + 35
        l_bsrn = qcr.REF.SOLAR_CONSTANT * 0.95 * math.pow(math.cos(math.radians(SZA)), 1.2) + 50

        return l1, l2, l_bsrn

    @staticmethod
    def lab(Dif, SZA, coef: Coef):
        return qcr.QC2(Dif, SZA, coef)

    def calc_lim(self, df, coef: Coef):
        return gen_calc_lim(self, df, coef)


class QC3:

    def __init__(self):
        self.name = 'QC3'
        self.varx = 'SZA'
        self.vary = 'direct_avg'
        self.unitx = 'SZA'
        self.unity = 'W.m-2'
        self.coefficients = {'level_1': 'C3', 'level_2': 'D3'}
        self.coef_range = [0.0, 1.0]

    @staticmethod
    def f(DirN, SZA, coef: Coef):
        ''' Return the 2 main variables, the 1rst level limit, the 2nd level limit and the physical limit for a sample'''

        l1 = qcr.REF.SOLAR_CONSTANT * coef.C3 * math.pow(math.cos(math.radians(SZA)), 0.2) + 10
        l2 = qcr.REF.SOLAR_CONSTANT * coef.D3 * math.pow(math.cos(math.radians(SZA)), 0.2) + 15
        l_bsrn = qcr.REF.SOLAR_CONSTANT

        return l1, l2, l_bsrn

    @staticmethod
    def lab(DirN, SZA, coef: Coef):
        return qcr.QC3(DirN, SZA, coef)

    def calc_lim(self, df, coef: Coef):
        return gen_calc_lim(self, df, coef)


class QC5:

    def __init__(self):
        self.name = 'QC5'
        self.varx = 'SZA'
        self.vary = 'downward_avg'
        self.unitx = 'SZA'
        self.unity = 'W.m-2'
        self.coefficients = {'level_1_min': 'C5', 'level_2_min': 'D5', 'level_1': 'C6', 'level_2': 'D6'}
        self.coef_range = [400.0, 600.0]
        self.coef_range_min = [200.0, 400.0]

    @staticmethod
    def f(LWdn, SZA, coef: Coef):
        ''' Return the 2 main variables, the 1rst level limit, the 2nd level limit and the physical limit for a sample'''

        l1_max = coef.C6
        l2_max = coef.D6
        l_bsrn_max = 700

        l1_min = coef.C5
        l2_min = coef.D5
        l_bsrn_min = 40

        return l1_max, l2_max, l_bsrn_max, l1_min, l2_min, l_bsrn_min

    @staticmethod
    def lab(LWdn, SZA, coef: Coef):
        return qcr.QC5(LWdn, coef)

    def calc_lim(self, df, coef: Coef):
        return gen_calc_lim(self, df, coef)


class QC10:

    def __init__(self):
        self.name = 'QC10'
        self.varx = 'temperature'
        self.vary = 'downward_avg'
        self.unitx = '°K'
        self.unity = 'W.m-2'
        self.coefficients = {'level_1_min': 'C11', 'level_2_min': 'D11', 'level_1': 'C12', 'level_2': 'D12'}
        self.coef_range = [0.0, 10.0]
        self.coef_range_min = [0.60, 1.0]

    @staticmethod
    def f(LWdn, Ta, coef: Coef):
        ''' Return the 2 main variables, the 1rst level limit, the 2nd level limit and the physical limit for a sample'''

        l1_max = qcr.REF.BOLTZMANN * math.pow(Ta + 273.15, 4) + coef.C12
        l2_max = qcr.REF.BOLTZMANN * math.pow(Ta + 273.15, 4) + coef.D12
        l_bsrn_max = 700

        l1_min = coef.C11 * qcr.REF.BOLTZMANN * math.pow(Ta + 274.15, 4)
        l2_min = coef.D11 * qcr.REF.BOLTZMANN * math.pow(Ta + 274.15, 4)
        l_bsrn_min = 40

        return l1_max, l2_max, l_bsrn_max, l1_min, l2_min, l_bsrn_min

    @staticmethod
    def lab(LWdn, Ta, coef: Coef):
        return qcr.QC10(LWdn, Ta + 273.15, coef)

    def calc_lim(self, df, coef: Coef):
        return gen_calc_lim(self, df, coef)



import math
from pybsrnqc.config import Coef

default_coef = Coef()


class REF:
    """BSRN Various quantities related to measurements"""
    TSWN = None  # temperature limit for albedo limit test, temp at which "snow" limit is used
    SOLAR_CONSTANT = 1368  # solar constant at mean Earth-Sun distance uses by QCRad in W.m-2
    BOLTZMANN = 5.67e-8  # Stephan-Boltzman constant = 5.67 x 10 -8 Wm -2 K -4


"""
Flag Value: Related to Type test:
5-6 Global Physical Limits (PP)
3-4 User configurable (UC2) 2nd level tests, also LW Tc and Td tests
1-2 User configurable (UC1) 1st level tests and non-definitive tests
0   No test failures
-1  Missing data or test not possible
"""


''' *********** QC1-QC6 (GSW, Diffuse SW, Direct SW, SWup, LWdn and LWup basic limits tests) *********** '''


def QC1(GSW, SZA, coef: Coef = default_coef):
    """GSW [basic limits tests]"""
    if GSW is not None:
        if (GSW < -4):
            return 5
        if (GSW < -2):
            return 3
        if SZA >= 0 and SZA <= 90: # cannot have float exponent of negative value
            if(GSW > (REF.SOLAR_CONSTANT * 1.5 * math.pow(math.cos(math.radians(SZA)), 1.2)) + 100):
                return 6
            if(GSW > (REF.SOLAR_CONSTANT * coef.D1 * math.pow(math.cos(math.radians(SZA)), 1.2)) + 55):
                return 4
            if(GSW > (REF.SOLAR_CONSTANT * coef.C1 * math.pow(math.cos(math.radians(SZA)), 1.2)) + 50):
                return 2
        else:
            return -1
    else:
        return -1
    return 0


def QC2(Dif, SZA, coef: Coef = default_coef):
    """Diffuse SW [basic limits tests]"""
    if Dif is not None:
        if (Dif < -4):
            return 5
        if(Dif < -2):
            return 3
        if SZA >= 0 and SZA <= 90: # cannot have float exponent of negative value
            if(Dif > (REF.SOLAR_CONSTANT * 0.95 * math.pow(math.cos(math.radians(SZA)), 1.2)) + 50):
                return 6
            if(Dif > (REF.SOLAR_CONSTANT * coef.D2 * math.pow(math.cos(math.radians(SZA)), 1.2)) + 35):
                return 4
            if(Dif > (REF.SOLAR_CONSTANT * coef.C2 * math.pow(math.cos(math.radians(SZA)), 1.2)) + 30):
                return 2
        else:
            return -1
    else:
        return -1
    return 0


def QC3(DirN, SZA, coef: Coef = default_coef):
    """Direct Normal SW  [basic limits tests]"""
    if DirN is not None:
        if (DirN < -4):
            return 5
        if(DirN < -2):
            return 3
        if (DirN > REF.SOLAR_CONSTANT):
            return 6
        if SZA >= 0 and SZA <= 90: # cannot have float exponent of negative value
            if(DirN > (REF.SOLAR_CONSTANT * coef.D3 * math.pow(math.cos(math.radians(SZA)), 0.2)) + 15):
                return 4
            if(DirN > (REF.SOLAR_CONSTANT * coef.C3 * math.pow(math.cos(math.radians(SZA)), 0.2)) + 10):
                return 2
        else:
            return -1
    else:
        return -1
    return 0


def QC4(SWup, SZA, coef: Coef = default_coef):
    """SWup [basic limits tests]"""
    if SWup is not None:
        if (SWup < -4):
            return 5
        if(SWup < -2):
            return 3
        if SZA >= 0 and SZA <= 90: # cannot have float exponent of negative value
            if (SWup > (REF.SOLAR_CONSTANT * 1.2 * math.pow(math.cos(math.radians(SZA)), 1.2)) + 50):
                return 6
            if(SWup > (REF.SOLAR_CONSTANT * coef.D4 * math.pow(math.cos(math.radians(SZA)), 1.2)) + 55):
                return 4
            if(SWup > (REF.SOLAR_CONSTANT * coef.C4 * math.pow(math.cos(math.radians(SZA)), 1.2)) + 50):
                return 2
        else:
            return -1
    else:
        return -1
    return 0


def QC5(LWdn, coef: Coef = default_coef):
    """LWdn [basic limits tests]"""
    if LWdn is not None:
        if (LWdn > 700):
            return 6
        if (LWdn < 40):
            return 5
        if (LWdn > coef.D6):
            return 4
        if (LWdn < coef.D5):
            return 3
        if (LWdn > coef.C6):
            return 2
        if (LWdn < coef.C5):
            return 1
    else:
        return -1
    return 0


def QC6(LWup, coef: Coef = default_coef):
    """LWup [basic limits tests]"""
    if LWup is not None:
        if (LWup > 900):
            return 6
        if (LWup < 40):
            return 5
        if (LWup > coef.D7):
            return 4
        if (LWup < coef.D8):
            return 3
        if (LWup > coef.C7):
            return 2
        if (LWup < coef.C8):
            return 1
    else:
        return -1
    return 0


def QC7(GSW, Dif, DirN, SZA):
    """GSW/Sum test [non-definitive]"""
    if all(v is not None for v in [GSW, Dif, DirN, SZA]):
        if SZA < 93:
            SumSW = Dif + (DirN * math.cos(math.radians(SZA)) )
            if SumSW > 50:
                if (SZA > 75 and SZA < 93 and ( ((GSW/SumSW)<0.85) or ((GSW/SumSW)>1.15) )):
                    return 2
                if (SZA < 75 and ( ((GSW/SumSW)<0.92) or ((GSW/SumSW)>1.08) )):
                    return 1
            else:
                return -1
        else:
            return -1
    else:
        return -1
    return 0


def QC8(Dif, GSW, SZA):
    """Dif/GSW test [non-definitive]"""
    if all(v is not None for v in [GSW, Dif, SZA]):
        if SZA >= 0 and SZA <= 90: # test between 0 and 90 degrees because we cannot have float exponent of negative value
            if GSW > 50:
                if (SZA > 75 and SZA < 93 and (Dif/GSW) > 1.10):
                    return 2
                if (SZA < 75 and (Dif/GSW) > 1.05):
                    return 1
            else:
                return -1
        else:
            return -1
    else:
        return -1
    return 0


def QC9(SWup, Dif, DirN, GSW, Ta, SZA, coef: Coef = default_coef):
    """SWup vs Sum SW test"""
    if all(v is not None for v in [GSW, SWup, Ta]):
        if all(v is not None for v in [Dif, DirN]):
            SumSW = Dif + (DirN * math.cos(math.radians(SZA)) )
            if (SumSW > 50 and GSW > 50 and SWup > SumSW and SWup > GSW):
                return 5
            if (SumSW > 50 and SWup > SumSW):
                return 3
            if ( Ta < REF.TSWN and (SumSW > 50 or GSW > 50) and SWup > (coef.C10*SumSW+25)):
                return 2
            if ( Ta >= REF.TSWN and (SumSW > 50 or GSW > 50) and SWup > (coef.C9*SumSW+25)):
                return 1
        elif (GSW > 50 and SWup > GSW):
                return 4
    else:
        return -1
    return 0


def QC10(LWdn, Ta, coef: Coef = default_coef):
    """LWdn to Ta test"""
    if all(v is not None for v in [LWdn, Ta]):
        if QC19(Ta) == 0:
            if (LWdn > (REF.BOLTZMANN * math.pow(Ta, 4) + coef.D12)):
                return 4
            if (LWdn < (coef.D11 * REF.BOLTZMANN * math.pow(Ta, 4))):
                return 3
            if (LWdn > (REF.BOLTZMANN * math.pow(Ta, 4) + coef.C12)):
                return 2
            if (LWdn < (coef.C11 * REF.BOLTZMANN * math.pow(Ta, 4))):
                return 1
        else:
            return -1
    else:
        return -1
    return 0


def QC11(LWup, Ta, coef: Coef = default_coef):
    """LWup to Ta test"""
    if all(v is not None for v in [LWup, Ta]):
        if QC19(Ta) == 0:
            if (LWup > (REF.BOLTZMANN * math.pow(Ta+coef.D14, 4))):
                return 4
            if (LWup < (REF.BOLTZMANN * math.pow(Ta-coef.D13, 4))):
                return 3
            if (LWup > (REF.BOLTZMANN * math.pow(Ta+coef.C14, 4))):
                return 2
            if (LWup < (REF.BOLTZMANN * math.pow(Ta-coef.C13, 4))):
                return 1
        else:
            return -1
    else:
        return -1
    return 0


def QC12(LWdn, LWup, coef: Coef = default_coef):
    """LWdn to LWup test"""
    if all(v is not None for v in [LWdn, LWup]):
        if LWdn > LWup + coef.D16:
            return 4
        if LWdn < LWup - coef.D15:
            return 3
        if LWdn > LWup + coef.C16:
            return 2
        if LWdn < LWup - coef.C15:
            return 1
    else:
        return -1
    return 0


def QC13(Tc, Ta, coef: Coef = default_coef):
    """LWdn Tc vs Ta"""
    if all(v is not None for v in [Tc, Ta]):
        if Tc > Ta + coef.C17D:
            return 4
        if Tc < Ta - coef.C17D:
            return 3
    else:
        return -1
    return 0


def QC14(Td, Ta, coef: Coef = default_coef):
    """LWdn Td vs Ta"""
    if all(v is not None for v in [Td, Ta]):
        if Td > Ta + coef.C17D:
            return 4
        if Td < Ta - coef.C17D:
            return 3
    else:
        return -1
    return 0


def QC15(Tc, Ta):
    """LWup Tc vs Ta"""
    # Do not have Tc from LWup
    return -1


def QC16(Td, Ta):
    """LWup Td vs Ta"""
    # Do not have Td from LWup
    return -1


def QC17(Tc, Td, coef: Coef = default_coef):
    """LWdn Tc vs Td"""
    if all(v is not None for v in [Tc, Td]):
        if (Tc - Td) > coef.C19:
            return 4
        if (Tc - Td) < coef.C18:
            return 3
    else:
        return -1
    return 0


def QC18(Tc, Td):
    """LWup Tc vs Td"""
    # Do not have Tc and Td from LWup
    return -1


def QC19(Ta):
    """Ta testing"""
    if Ta is not None:
        if Ta > 350 or Ta < 170:
            return 1
    else:
        return -1
    return 0



"""
Module that compute the coefficients regarding the historical data
 """

import importlib.resources
import json

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score
from pybsrnqc.config import Coef 

# -----------------------------------------------------------------------------------------------------------
# Initialisation of a set of coefficients
# Get data conf from JSON file

dic_coefs = Coef()
# ------------------------------------------------------------------------------------------------------------


def coef_variation(df, log_kernel, QC, level='level_2', coef_range=[0.0, 1.2], step=0.01, coef_range_min=None, step_min=None, verbose=True, display=True):

    """ Function which calculates certain indicators
    according to the value of the coefficient """

    # Creation of the dataframe containing the KDE values
    df['log_kde'] = log_kernel
    df_min = df.copy()
    df_min = df_min.loc[df_min['downward_avg'] <= 400]

    # Name of the coefficient we study
    coef_name = QC.coefficients[level]

    # Values of the coefficients we study
    values_c = list(np.arange(coef_range[0], coef_range[1], step))

    # Initialisation of the indicators
    the_coefs = []
    nb_out = []
    kde_means = []
    kde_sums = []
    kde_stds = []

    if QC.vary == 'downward_avg':

        coef_name_min = QC.coefficients[level + '_min']
        values_c_min = list(np.arange(coef_range_min[0], coef_range_min[1], step_min))

        # Initialisation of the indicators
        the_coefs_min = []
        nb_out_min = []
        kde_means_min = []
        kde_sums_min = []
        kde_stds_min = []

    # Visualization of the computation
    if verbose:
        for k in range(len(values_c)):
            print('.', end='')
        print()

    # Calculus of the indicators for each coefficients
    for coef in values_c:

        # Generation of the new coefficients

        dic = Coef()
        dic.__setattr__(coef_name, coef)

        # Generation of the outliers according to the limits linked to the coefficients

        df_lim = df.copy()
        df_lim['flag'] = QC.calc_lim(df, coef=dic)[3]
        df_lim['out_coef'] = np.array([0] * df_lim.shape[0])
        df_lim.loc[df_lim.flag == 4, "out_coef"] = 1

        df_lim_out = df_lim.loc[df_lim['out_coef'] == 1]
        kde_mean = df_lim_out['log_kde'].mean()
        kde_sum = df_lim_out['log_kde'].sum()
        kde_std = df_lim_out['log_kde'].std()

        nb_out.append(df_lim['out_coef'].sum())
        kde_means.append(kde_mean)
        kde_sums.append(kde_sum)
        the_coefs.append(coef)
        kde_stds.append(kde_std)

        # Visualization of the computation
        print('-', end='')

    if QC.vary == 'downward_avg':
        for coef_min in values_c_min:

            # Generation of the new coefficients
            dic = Coef()
            dic.__setattr__(coef_name_min, coef_min)

            # Generation of the outliers according to the limits linked to the coefficients
            df_lim = df_min.copy()
            df_lim['flag'] = QC.calc_lim(df_min, coef=dic)[3]
            df_lim['out_coef'] = np.array([0] * df_lim.shape[0])
            df_lim.loc[df_lim.flag == 3, "out_coef"] = 1

            df_lim_out = df_lim.loc[df_lim['out_coef'] == 1]
            kde_mean = df_lim_out['log_kde'].mean()
            kde_sum = df_lim_out['log_kde'].sum()
            kde_std = df_lim_out['log_kde'].std()

            nb_out_min.append(df_lim['out_coef'].sum())
            kde_means_min.append(kde_mean)
            kde_sums_min.append(kde_sum)
            the_coefs_min.append(coef_min)
            kde_stds_min.append(kde_std)

            # Visualization of the computation
            print('-', end='')

    tab = np.zeros((len(values_c), 4))
    df_var = pd.DataFrame(tab)
    df_var['nb_out'] = nb_out
    df_var['kde_mean'] = kde_means
    df_var['kde_sum'] = kde_sums
    df_var['coefs'] = the_coefs
    df_var['std'] = kde_stds
    df_var = df_var.drop([0, 1, 2, 3], axis=1)

    # Display the plot
    if display:

        plt.figure(figsize=(20, 14))

        # Number of values removed
        ax1 = plt.subplot(221)
        ax1.plot(df_var['coefs'], df_var['nb_out'])
        ax1.set_xlabel('Coefficients')
        ax1.set_ylabel('Number of values removed')
        ax1.set_title('Number of outlier according to the coefficient')

        # KDE Mean
        ax2 = plt.subplot(222)
        ax2.plot(df_var['coefs'], df_var['kde_mean'])
        ax2.set_xlabel('Coefficients')
        ax2.set_ylabel('KDE Mean')
        ax2.set_title('Kde mean of the outliers according to the coefficient')

        # KDE Mean vs nb_out
        ax3 = plt.subplot(223)
        ax3.plot(df_var['nb_out'], df_var['kde_mean'])
        ax3.set_xlabel('Number of values removed')
        ax3.set_ylabel('KDE Mean')
        ax3.set_title('Kde mean of the outliers versus number of values removed')

        # KDE std versus coefs
        ax4 = plt.subplot(224)
        ax4.plot(df_var['coefs'], df_var['std'])
        ax4.set_xlabel('Coefficients')
        ax4.set_ylabel('KDE std')
        ax4.set_title('Kde standard deviation according to the coefficients')

        plt.show()

    if QC.vary == 'downward_avg':

        tab = np.zeros((len(values_c_min), 4))
        df_var_min = pd.DataFrame(tab)
        df_var_min['nb_out'] = nb_out_min
        df_var_min['kde_mean'] = kde_means_min
        df_var_min['kde_sum'] = kde_sums_min
        df_var_min['coefs'] = the_coefs_min
        df_var_min['std'] = kde_stds_min
        df_var_min = df_var_min.drop([0, 1, 2, 3], axis=1)

        # Display the plot
        if display:

            plt.figure(figsize=(20, 14))

            # Number of values removed
            ax1 = plt.subplot(221)
            ax1.plot(df_var_min['coefs'], df_var_min['nb_out'])
            ax1.set_xlabel('Coefficients')
            ax1.set_ylabel('Number of values removed')
            ax1.set_title('Number of outlier according to the coefficient')

            # KDE Mean
            ax2 = plt.subplot(222)
            ax2.plot(df_var_min['coefs'], df_var_min['kde_mean'])
            ax2.set_xlabel('Coefficients')
            ax2.set_ylabel('KDE Mean')
            ax2.set_title('Kde mean of the outliers according to the coefficient')

            # KDE Mean vs nb_out
            ax3 = plt.subplot(223)
            ax3.plot(df_var_min['nb_out'], df_var_min['kde_mean'])
            ax3.set_xlabel('Number of values removed')
            ax3.set_ylabel('KDE Mean')
            ax3.set_title('Kde mean of the outliers versus number of values removed')

            # KDE std versus coef
            ax4 = plt.subplot(224)
            ax4.plot(df_var_min['coefs'], df_var_min['std'])
            ax4.set_xlabel('Coefficients')
            ax4.set_ylabel('KDE std')
            ax4.set_title('Kde standard deviation according to the coefficients')

            plt.title('Indicators relative to the minimum limits')
            plt.show()

            return df_var, df_var_min

    return df_var


def threshold_var(df, log_kernel, threshold_range=None, step=0.1, display=True):

    if threshold_range is None:

        threshold_range = [np.min(log_kernel), np.max(log_kernel)]

    values_t = list(np.arange(threshold_range[0], threshold_range[1], step))

    # Gettinf KDE density
    df['log_kde'] = log_kernel

    # Initialisation
    nb_out_density = []

    for threshold in values_t:

        df['out_density'] = np.array([0] * df.shape[0])
        df['out_density_min'] = np.array([0] * df.shape[0])

        # Labeling according to the chosen threshold
        df.loc[df.log_kde <= threshold, "out_density"] = 1
        df_min = df.copy()
        df_min.loc[(df['downward_avg'] <= 350) & (df.log_kde <= threshold), 'out_density_min'] = 1

        nb_out_density.append(df['out_density'].sum())

    tab = np.zeros((len(values_t), 1))
    df_var = pd.DataFrame(tab)
    df_var['nb_out_density'] = nb_out_density
    df_var['thresholds'] = values_t
    df_var = df_var.drop([0], axis=1)

    if display:

        plt.figure(figsize=(20, 14))
        plt.plot(df_var['thresholds'], df_var['nb_out_density'])
        plt.title('Number of elements removed according to a threshold')
        plt.xlabel('Threshold')
        plt.ylabel('Removed values')
        plt.show()

    return df_var


# ---------------------------------------------------------------------------------------------------------------------

def calc_coef(df, log_kernel, QC, threshold, level='level_2', coef_range=[0.0, 1.2], coef_range_min=[0.0, 1.2], step=0.01, step_min=0.1, verbose=True, selected=None):

    # Labeling according to the chosen threshold
    df['log_kde'] = log_kernel
    df['out_density'] = np.array([0] * df.shape[0])
    df['out_density_min'] = np.array([0] * df.shape[0])
    df.loc[df.log_kde <= threshold, "out_density"] = 1
    df.loc[(df['downward_avg'] <= 350) & (df.log_kde <= threshold), 'out_density_min'] = 1

    # Values of the coefficients tested
    values_c = list(np.arange(coef_range[0], coef_range[1], step))

    # Name of the coefficient we study
    coef_name = QC.coefficients[level]

    # Adding a constraint for QC3 and QC10
    if QC.name == 'QC3' or QC.name == 'QC10':
        # We precise that points under a certain limit are not ouliers
        dic = Coef()
        dic.__setattr__(coef_name, 0.6)

        for el in selected:
            df.loc[(df['temperature'] == el[0]) & (df[QC.vary] == el[1]), 'out_density'] = 0

    # Score initialisation
    a_scores = []
    p_scores = []
    r_scores = []
    f_scores = []

    if QC.vary == 'downward_avg':
        coef_name_min = QC.coefficients[level + '_min']
        values_c_min = list(np.arange(coef_range_min[0], coef_range_min[1], step_min))
        a_scores_min = []
        p_scores_min = []
        r_scores_min = []
        f_scores_min = []

    # Verbose
    if verbose:
        for k in range(len(values_c)):
            print('.', end='')
        print()

    for coef in values_c:

        if verbose:
            print('-', end='')

        # Generation of coefficient
        dic = Coef()
        dic.__setattr__(coef_name, coef)

        # Generation of the outliers according to the equations and the density

        df_lim = df.copy()
        df_lim['flag'] = QC.calc_lim(df, coef=dic)[3]

        if level == 'level_2':
            df_lim['out_coef'] = np.array([0] * df_lim.shape[0])
            df_lim.loc[df_lim.flag == 4, "out_coef"] = 1
        else:
            df_lim['out_coef'] = np.array([0] * df_lim.shape[0])
            df_lim.loc[(df_lim.flag == 2) | (df_lim.flag == 4), "out_coef"] = 1

        a_scores.append(accuracy_score(df['out_density'], df_lim['out_coef']))
        p_scores.append(precision_score(df['out_density'], df_lim['out_coef'], zero_division=0))
        r_scores.append(recall_score(df['out_density'], df_lim['out_coef'], zero_division=0))
        f_scores.append(f1_score(df['out_density'], df_lim['out_coef'], zero_division=0))

    if QC.vary == 'downward_avg':

        for coef_min in values_c_min:

            # Generation of coefficient

            dic = Coef()
            dic.__setattr__(coef_name_min, coef_min)

            # Generation of the outliers according to the equations and the density

            df_lim = df.copy()
            df_lim['flag'] = QC.calc_lim(df, coef=dic)[3]

            if level == 'level_2':
                df_lim['out_coef'] = np.array([0] * df_lim.shape[0])
                df_lim.loc[df_lim.flag == 3, "out_coef"] = 1
            else:
                df_lim['out_coef'] = np.array([0] * df_lim.shape[0])
                df_lim.loc[(df_lim.flag == 1) | (df_lim.flag == 3), "out_coef"] = 1

            a_scores_min.append(accuracy_score(df['out_density'],
                                df_lim['out_coef']))
            p_scores_min.append(precision_score(df['out_density'], df_lim['out_coef'], zero_division=0))
            r_scores_min.append(recall_score(df['out_density'], df_lim['out_coef'], zero_division=0))
            f_scores_min.append(f1_score(df['out_density'], df_lim['out_coef'], zero_division=0))

    # Creation of the scores dataframe

    tab = np.zeros((len(values_c), 4))
    df_scores = pd.DataFrame(tab)
    df_scores[coef_name] = values_c
    df_scores['accuracy_score'] = a_scores
    df_scores['precision_score'] = p_scores
    df_scores['recall_score'] = r_scores
    df_scores['f1_score'] = f_scores
    df_scores = df_scores.drop([0, 1, 2, 3], axis=1)

    # Getting the coefficient with the best score
    f1_max = df_scores['f1_score'].max()
    line_max = df_scores.loc[df_scores['f1_score'] == f1_max]

    if QC.vary == 'downward_avg':

        tab = np.zeros((len(values_c_min), 4))
        df_scores_min = pd.DataFrame(tab)

        df_scores_min[coef_name_min] = values_c_min
        df_scores_min['accuracy_score'] = a_scores_min
        df_scores_min['precision_score'] = p_scores_min
        df_scores_min['recall_score'] = r_scores_min
        df_scores_min['f1_score'] = f_scores_min
        df_scores_min = df_scores_min.drop([0, 1, 2, 3], axis=1)

        # Getting the coefficient with the best score
        f1_max_min = df_scores_min['f1_score'].max()
        line_max_min = df_scores_min.loc[df_scores_min['f1_score'] == f1_max_min]

    # Print the coefficient
    print()
    print('Upper limit')
    print(line_max)

    if QC.vary == 'downward_avg':
        print()
        print('Lower limit')
        print(line_max_min)

        return df_scores, line_max, df_scores_min, line_max_min

    return df_scores, line_max



"""
Module that plot the curves following the quality controle equations
 """
import os
import pickle

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.path import Path
from matplotlib.widgets import LassoSelector
from scipy import stats

from pybsrnqc.config import Coef

# -----------------------------------------------------------------------------------------------------------
# Plots of the BSRN limits on datasets


def limit_plot(df, QC, coef: Coef, save: bool = False, level='all', display: bool = True,
               values: bool = True, fig: bool = True):

    """ Function plotting the limit curves and the dataset points of a
        dataframe"""

    df = df[['timestamp', QC.varx, QC.vary]]

    X_val1 = np.array(df[QC.varx])
    X_val2 = np.array(df[QC.vary])

    lim_l1 = []
    lim_l2 = []
    lim_bsrn = []

    lim_l1_min = []
    lim_l2_min = []
    lim_bsrn_min = []

    val1_var = list(X_val1)
    val2_var = list(X_val2)

    for VAL1, VAL2 in zip(X_val1, X_val2):

        if QC.vary == 'downward_avg':

            l1, l2, l_bsrn, l1_min, l2_min, l_bsrn_min = QC.f(VAL2, VAL1, coef)

            lim_l1_min.append(l1_min)
            lim_l2_min.append(l2_min)
            lim_bsrn_min.append(l_bsrn_min)

        else:

            l1, l2, l_bsrn = QC.f(VAL2, VAL1, coef)

        lim_l1.append(l1)
        lim_l2.append(l2)
        lim_bsrn.append(l_bsrn)

    if fig:
        plt.figure(figsize=(20, 14))

    if values:
        plt.scatter(val1_var, val2_var, marker='+', color='dodgerblue',
                    label='Values', s=1)

    if level == 'level_1' or level == 'all':
        plt.scatter(val1_var, lim_l1, marker='+', color='green',
                    label=f"1rst level limit - {coef.__getattribute__(QC.coefficients['level_1'])}", s=1)
    if level == 'level_2' or level == 'all':
        plt.scatter(val1_var, lim_l2, marker='+', color='blue',
                    label=f"2nd level limit - {coef.__getattribute__(QC.coefficients['level_2'])}", s=1)
    if level == 'level_bsrn' or level == 'all':
        plt.scatter(val1_var, lim_bsrn, marker='+', color='red', label="BSRN  limit", s=1)

    if QC.vary == 'downward_avg':

        if level == 'level_1' or level == 'all':
            plt.scatter(val1_var, lim_l1_min, marker='+', color='lightgreen',
                        label=f"1rst level limit - {coef.__getattribute__(QC.coefficients['level_1_min'])}", s=1)
        if level == 'level_2' or level == 'all':
            plt.scatter(val1_var, lim_l2_min, marker='+', color='lightblue',
                        label=f"2nd level limit - {coef.__getattribute__(QC.coefficients['level_2_min'])}", s=1)
        if level == 'level_bsrn' or level == 'all':
            plt.scatter(val1_var, lim_bsrn_min, marker='+', color='lightcoral', label="BSRN  limit", s=1)

    if save:
        plt.savefig(f'limit_plot - {QC.vary} versus {QC.varx}')
    if display:
        plt.xlabel(QC.unitx,)
        plt.ylabel(QC.unity)
        plt.title(f'Limits - {QC.vary}', fontsize=20)
        plt.legend(prop={'size': 15})
        plt.show()


def multiplot_coef(df, QC, coef, level='level_2', coef_values=[0.0, 0.5, 1.2], level_min=None, coef_values_min=None):
    'Dessine la courbe limite pour plusieurs valeurs du coefficients'

    X_val1 = np.array(df[QC.varx])
    X_val2 = np.array(df[QC.vary])

    val1_var = list(X_val1)
    val2_var = list(X_val2)

    plt.figure(figsize=(20, 14))
    plt.scatter(val1_var, val2_var, marker='+', color='dodgerblue',
                label='Values', s=1)

    for v in coef_values:

        # We compute the considered coefficient
        coef.__setattr__(QC.coefficients[level], v)

        # We plot for this coefficient
        limit_plot(df, QC, coef, level=level, display=False, values=False, fig=False)

    if coef_values_min is not None:
        for v in coef_values_min:

            # We compute the considered coefficient
            coef.__setattr__(QC.coefficients[level_min], v)

            # We plot for this coefficient
            limit_plot(df, QC, coef, level=level, display=False, values=False, fig=False)

    plt.xlabel(QC.unitx)
    plt.ylabel(QC.unity)
    plt.title(f'Limits - {QC.vary}', fontsize=20)
    plt.legend(prop={'size': 15})
    plt.show()


# ----------------------------------------------------------------------------------------------------------------------
# Displaying of the data histograms in 2D and 3D. It allows us to see the
# shape of the data and its caracteristics like unimodality

def hist_data(df, QC, dimension='3D'):

    # Get the data
    x = np.array(df[[QC.vary, QC.varx]])

    # Create the figure
    fig = plt.figure(figsize=(20, 14))
    ax = fig.add_subplot(projection='3d')

    hist, xedges, yedges = np.histogram2d(x[:, 0], x[:, 1], bins=75)

    # Construct arrays for the anchor positions of the 16 bars.
    xpos, ypos = np.meshgrid(xedges[:-1] + 0.25, yedges[:-1] + 0.25, indexing="ij")
    xpos = xpos.ravel()
    ypos = ypos.ravel()
    zpos = 0

    # Construct arrays with the dimensions for the 16 bars.
    dx = xedges[-1] / 75 * np.ones_like(zpos)
    dy = yedges[-1] / 75 * np.ones_like(zpos)
    dz = hist.ravel()

    cmap = plt.cm.get_cmap('Wistia')  # Get desired colormap - you can change this
    max_height = np.max(dz)   # get range of colorbars so we can normalize
    min_height = np.min(dz)
    # scale each z to [0,1], and get their rgb values
    rgba = [cmap((k - min_height) / max_height) for k in dz]

    ax.bar3d(xpos, ypos, zpos, dx, dy, dz, color=rgba, zsort='average')

    plt.title(f'Data histogram - {QC.vary} vs {QC.varx}', fontdict={'fontsize': 20})
    plt.xlabel(QC.unity)
    plt.ylabel(QC.unitx)
    ax.set_zlabel('Number of samples')
    plt.show()


# --------------------------------------------------------------------------------------------
# KDE computation and plotting for our dataset


def kde_computing(df, QC, display=True, coef: Coef = None, limits=False, level='All',
                  log_form=True, save='KDE_result', select=False, bw_sel=None):

    # Get the data
    X = np.array(df[[QC.vary, QC.varx]]).T
    # Kernel calculation
    print('Computing kde - It can take some times')

    kernel = stats.gaussian_kde(X, bw_method=bw_sel)(X)
    kernel_log = np.log(kernel)

    # Plot
    if display:

        fig, ax = plt.subplots(figsize=(20, 14))
        pts = plt.scatter(X[1, :], X[0, :], c=kernel_log, s=1, cmap=plt.cm.jet)
        selector = SelectFromCollection(ax, pts)

    if limits:
        limit_plot(df, QC, coef, save=False, level='all', display=False, values=False, fig=False)

    if display:
        if select:
            if QC.name == 'QC3' or QC.name == 'QC10':

                def accept(event):
                    if event.key == "enter":
                        # print("Selected points:")
                        selected = selector.xys[selector.ind].tolist()
                        with open("selected.txt", "wb") as fp:   # Pickling
                            pickle.dump(selected, fp)
                        # print(selected)
                        # print(type(selected))
                        selector.disconnect()
                        ax.set_title("", fontsize=20)
                        fig.canvas.draw()

                fig.canvas.mpl_connect("key_press_event", accept)
            ax.set_title("Press enter to accept selected points.", fontsize=20)

        ax.set_xlabel(QC.unitx)
        ax.set_ylabel(QC.unity)
        if select:
            if QC.name == 'QC3' or QC.name == 'QC10':
                ax.set_title(f"{QC.vary} density along {QC.varx} \n Select the points that you don't want to consider", fontsize=20)
        else:
            ax.set_title(f'{QC.vary} density along {QC.varx}', fontsize=20)
        plt.colorbar(pts, label='Density Log(KDE)')
        ax.legend(prop={'size': 15})
        plt.show()

    # Download file
    if save is not None:
        pd.DataFrame(kernel).to_csv(save + '.csv')

    # Return density
    if log_form:
        if select:
            if QC.name == 'QC3' or QC.name == 'QC10':
                with open("selected.txt", "rb") as fp:   # Unpickling
                    selected = pickle.load(fp)
                os.remove("selected.txt")
                return kernel_log, selected
        else:
            return kernel_log
    else:
        if select:
            if QC.name == 'QC3' or QC.name == 'QC10':
                with open("selected.txt", "rb") as fp:   # Unpickling
                    selected = pickle.load(fp)
                os.remove("selected.txt")
                return kernel, selected
        else:
            return kernel


def plot_kde(df, log_kernel, QC, coef: Coef, level='level_2'):

    plt.figure(figsize=(20, 14))

    # Get the data
    X = np.array(df[[QC.vary, QC.varx]]).T

    # Plot the data
    plot = plt.scatter(X[1, :], X[0, :], c=log_kernel, s=1, cmap=plt.cm.jet)

    if level is not None:
        limit_plot(df, QC, coef, save=False, level=level, display=False, values=False, fig=False)

    plt.xlabel(QC.unitx)
    plt.ylabel(QC.unity)
    plt.title(f'{QC.vary} density along {QC.varx}', fontsize=20)
    plt.colorbar(plot, label='Density Log(KDE)')
    plt.legend(prop={'size': 15})
    plt.show()


def plot_series_kde(df, log_kernel, QC, begin, end, var='timestamp', line=True):

    df['log_kde'] = log_kernel
    kernel_range = [np.min(log_kernel), np.max(log_kernel)]
    mask = (df['timestamp'] >= begin) & (df['timestamp'] < end)
    df_week = df.loc[mask]

    if var == 'SZA':
        tick = None
    else:
        tick = np.arange(0, len(list(df_week[var])), 60)

    plt.figure(figsize=(20, 14))

    if line:
        plt.plot(list(df_week[var]), list(df_week[QC.vary]),
                 color='black', alpha=0.2)

    plot = plt.scatter(list(df_week[var]), list(df_week[QC.vary]),
                       c=np.array(df_week['log_kde']), cmap=plt.cm.jet, s=5)

    plt.xlabel(var)
    plt.ylabel(QC.unity)
    plt.xticks(tick, rotation=45)
    plt.title(f'{QC.vary} and its density - {begin}', fontsize=20)
    plt.clim(kernel_range[0], kernel_range[1])
    plt.colorbar(plot, label='Density Log(KDE)')
    plt.show()


# -----------------------------------------------
# Lasso used to select the datas

class SelectFromCollection:
    """
    Select indices from a matplotlib collection using `LassoSelector`.

    Selected indices are saved in the `ind` attribute. This tool fades out the
    points that are not part of the selection (i.e., reduces their alpha
    values). If your collection has alpha < 1, this tool will permanently
    alter the alpha values.

    Note that this tool selects collection objects based on their *origins*
    (i.e., `offsets`).

    Parameters
    ----------
    ax : `~matplotlib.axes.Axes`
        Axes to interact with.
    collection : `matplotlib.collections.Collection` subclass
        Collection you want to select from.
    alpha_other : 0 <= float <= 1
        To highlight a selection, this tool sets all selected points to an
        alpha value of 1 and non-selected points to *alpha_other*.
    """

    def __init__(self, ax, collection, alpha_other=0.3):
        self.canvas = ax.figure.canvas
        self.collection = collection
        self.alpha_other = alpha_other

        self.xys = collection.get_offsets()
        self.Npts = len(self.xys)

        # Ensure that we have separate colors for each object
        self.fc = collection.get_facecolors()
        if len(self.fc) == 0:
            raise ValueError('Collection must have a facecolor')
        elif len(self.fc) == 1:
            self.fc = np.tile(self.fc, (self.Npts, 1))

        self.lasso = LassoSelector(ax, onselect=self.onselect)
        self.ind = []

    def onselect(self, verts):
        path = Path(verts)
        self.ind = np.nonzero(path.contains_points(self.xys))[0]
        self.fc[:, -1] = self.alpha_other
        self.fc[self.ind, -1] = 1
        self.collection.set_facecolors(self.fc)
        self.canvas.draw_idle()

    def disconnect(self):
        self.lasso.disconnect_events()
        self.fc[:, -1] = 1
        self.collection.set_facecolors(self.fc)
        self.canvas.draw_idle()
